.text	

.globl	_ossl_rsaz_amm52x40_x1_ifma256
.globl	_ossl_rsaz_amm52x40_x2_ifma256
.globl	_ossl_extract_multiplier_2x40_win5

_ossl_rsaz_amm52x40_x1_ifma256:
_ossl_rsaz_amm52x40_x2_ifma256:
_ossl_extract_multiplier_2x40_win5:
.byte	0x0f,0x0b
	.byte	0xf3,0xc3

